
#region Header

/* /////////////////////////////////////////////////////////////////////////////////////////////////
Pose Manager for Virt-A-Mate by Fabio Den.

Written in Notepad++
	
Version 1.0 2021-05-05
	Initial release.
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Animations;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;

#endregion

public class PoseManager : MVRScript
{	
	#region Variables
	
	protected UIDynamicPopup _displayPopup; // any UI element
	UIDynamic sep1;
	UIDynamic sep2;
	
	private JSONStorableStringChooser[] person = new JSONStorableStringChooser[3];		// people load order
	private JSONStorableStringChooser[] obj3ct = new JSONStorableStringChooser[3];		// object load order
	private AtomCallbacks atomCalls = new AtomCallbacks();
	public JSONClass jc = new JSONClass();
	
	protected JSONStorableBool deleteObjects;
	
	protected bool enablePhysics;
	UIDynamicButton enablePhys;
	UIDynamicButton disablePhys;
	
	protected JSONStorableBool spawnObjects;
	
	private bool creatingSubscene = false;
	private string subname = "rootPM";

	private UIDynamicTextField myLabel;
	
	private const string _saveExt = "json";
	private const string _saveFolder = "Saves/Person/pose/PoseManager/";
	
	#endregion

	#region Events

	public override void Init()
	{
		SetupGUI();
	}
	
	protected void OnSceneLoaded()
	{
		for(int i = 1; i <= person.Length-1; i++)
		{
			person[i].val = "";
		}
		
		for(int i = 1; i <= obj3ct.Length-1; i++)
		{
			obj3ct[i].val = "";
		}
		
		atomCalls.Clean(false);
	}
	
	public void OnEnable()
	{
		atomCalls.SetParent(this);
		SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
	}

	public void OnDisable()
	{
		SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
	}

	public void OnDestroy()
	{

	}
	
	#endregion

	#region Gui

	public void SetupGUI()
	{
		int sep = 5;
		CreateSpacer(false).height = sep; 
		CreateSpacer(true).height = sep;
		
		// load and save
		var btn = CreateButton("Load", true);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { Import(); } );

		btn = CreateButton("Save", false);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { Export(); } );
		
		myLabel = CreateTextField(new JSONStorableString("", ""), false);
		myLabel.text = "Remember to <b>select the people and objects</b> before attempting to make a save.";
		myLabel.height = 115;
		
		myLabel = CreateTextField(new JSONStorableString("", ""), true);
		myLabel.text = "Press <b>Install</b> to update the built-in pose library. Select <b>Load</b> to choose a new pose set.";
		myLabel.height = 115;
		
		btn = CreateButton("Install", false);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { Install(); } );
		
		deleteObjects = new JSONStorableBool("deleteObjects", true, SyncDelete);
		var toggle = CreateToggle(deleteObjects, true);
		toggle.label = "Clean up";
		RegisterBool(deleteObjects);
		
		CreateSpacer(false).height = sep; 
		CreateSpacer(true).height = sep;
	
		// center
		enablePhysics = true;	
		enablePhys = CreateButton("<b>Enable physics</b>", true);
		enablePhys.buttonColor = new Color(0.780f, 0.899f, 0.780f);
		enablePhys.button.onClick.AddListener( () => { RestorePhysics(); } );	
		
		disablePhys = CreateButton("Disable physics", true);
		disablePhys.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		disablePhys.button.onClick.AddListener( () => { DisablePhysics(); } );
		
		spawnObjects = new JSONStorableBool("spawnObjects", true, SyncSpawn);
		toggle = CreateToggle(spawnObjects, false);
		toggle.label = "Spawn Objects";
		RegisterBool(spawnObjects);
		
		btn = CreateButton("Move to Origin", false);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { MoveToOrigin(); } );
		
		btn = CreateButton("Move to Camera View", true);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { MoveToCamera(); } );
		
		btn = CreateButton("Move around", false);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { Move(); } );

		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;

		// selectors
		btn = CreateButton("Add object selector", false);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { AddStorableObject(); } );

		btn = CreateButton("Add person selector", true);
		btn.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		btn.button.onClick.AddListener( () => { AddStorablePerson(); } );			

		for(int i = 1; i <= person.Length-1; i++)
		{
			int x = i; // each index needs its own int
			bool isperson = true;
			person[i] = new JSONStorableStringChooser("person" + i.ToString(), null, "", "", (string atomUID) => SyncAtom(atomUID, x, isperson));
			RegisterStringChooser(person[i] );
			UIDynamicPopup dp = CreateFilterablePopup(person[i], true);
			dp.popupPanelHeight = 500f;
			dp.label = "Person #" + i.ToString();
			dp.popup.onOpenPopupHandlers += () => SyncAtomChoices(x, isperson);
		}
		
		for(int i = 1; i <= obj3ct.Length-1; i++)
		{
			int x = i; // each index needs its own int
			bool isperson = false;
			obj3ct[i] = new JSONStorableStringChooser("obj3ct" + i.ToString(), null, "", "", (string atomUID) => SyncAtom(atomUID, x, isperson));
			RegisterStringChooser(obj3ct[i] );
			UIDynamicPopup dp = CreateFilterablePopup(obj3ct[i], false);
			dp.popupPanelHeight = 500f;
			dp.label = "Object #" + i.ToString();
			dp.popup.onOpenPopupHandlers += () => SyncAtomChoices(x, isperson);
		}
		
		sep1 = CreateSpacer(false);
		sep1.height = 500.0f;
		sep2 = CreateSpacer(true);
		sep2.height = 500.0f;
	}
	
	private void SyncDelete(bool flag)
	{
		deleteObjects.val = flag;
	}
	
	private void SyncSpawn(bool flag)
	{
		spawnObjects.val = flag;
	}
	
	private void AddStorablePerson()
	{
		RemoveSpacer(sep2);
		
		int size = person.Length;	
		JSONStorableStringChooser[] temp = new JSONStorableStringChooser[size + 1];
		 
		for (int c = 1; c < size; c++ )
		{
				temp[c] = person[c];
		}
		
		person = temp;
		int i = size;
		 
		int x = i; // each index needs its own int
		bool isperson = true;
		person[i] = new JSONStorableStringChooser("person" + i.ToString(), null, "", "", (string atomUID) => SyncAtom(atomUID, x, isperson));
		RegisterStringChooser(person[i] );
		UIDynamicPopup dp = CreateFilterablePopup(person[i], true);
		dp.popupPanelHeight = 500f;
		dp.label = "Person #" + i.ToString();
		dp.popup.onOpenPopupHandlers += () => SyncAtomChoices(x, isperson);
		
		sep2 = CreateSpacer(true);
		sep2.height = 500.0f;
	}
	
	private void AddStorableObject()
	{
		RemoveSpacer(sep1);
		
		int size = obj3ct.Length;	
		JSONStorableStringChooser[] temp = new JSONStorableStringChooser[size + 1];
		 
		for (int c = 1; c < size; c++ )
		{
				temp[c] = obj3ct[c];
		}
		
		obj3ct = temp;
		int i = size;
		
		int x = i; // each index needs its own int
		bool isperson = false;
		obj3ct[i] = new JSONStorableStringChooser("obj3ct" + i.ToString(), null, "", "", (string atomUID) => SyncAtom(atomUID, x, isperson));
		RegisterStringChooser(obj3ct[i] );
		UIDynamicPopup dp = CreateFilterablePopup(obj3ct[i], false);
		dp.popupPanelHeight = 500f;
		dp.label = "Object #" + i.ToString();
		dp.popup.onOpenPopupHandlers += () => SyncAtomChoices(x, isperson);
		
		sep1 = CreateSpacer(false);
		sep1.height = 500.0f;
	}
	
	private void AddMissingPersons(int requiredpeople)
	{
		int size = person.Length;
		int max = requiredpeople - size + 1;
		
		if (max >= 1)
		{
			for (int c = 1; c <= max; c++ )
			{
					AddStorablePerson();
			}			
		}
	}
	
	private void AddMissingObjects(int requiredobjects) // ie: 4 objects
	{
		int size = obj3ct.Length; // 4 by default
		int max = requiredobjects - size + 1;
	
		if (max >= 1)
		{
			for (int c = 1; c <= max; c++ )
			{
					AddStorableObject();
			}			
		}
	}
	
	// before opening the popup
	private void SyncAtomChoices(int i, bool isperson)
	{
		List<string> mychoices = null;
		
		if (isperson == false)
			mychoices = SuperController.singleton.GetAtoms().Where(a => a.category != "People").Select(a => a.name).ToList();
		else
			mychoices = SuperController.singleton.GetAtoms().Where(a => a.category == "People").Select(a => a.name).ToList();
		
		mychoices.Remove("WindowCamera");
		mychoices.Remove("PlayerNavigationPanel");
		mychoices.Remove("CoreControl");
		mychoices.Remove("[CameraRig]");
		
		if (isperson == false)
			obj3ct[i].choices = mychoices;
		else
			person[i].choices = mychoices;
	}

	// after choosing a string from the popup
	private void SyncAtom(string atomUID, int i, bool isperson)
	{
		if (isperson == false)
			obj3ct[i].val = atomUID;
		else
			person[i].val = atomUID;
	}

	#endregion

	#region Install

	private void Install()
	{
		string oldDir = GetPluginPath(this) + "/Poses";
		string newDir = _saveFolder;
		int dirCount = 0;
		int fileCount = 0;
		int percent = 0;
		float increment = 0f;
		
		if ( FileManagerSecure.DirectoryExists(newDir) == false )
			FileManagerSecure.CreateDirectory(newDir);
		
		string[] dirs = FileManagerSecure.GetDirectories(oldDir);
		dirCount = dirs.Length;	
		increment = (100f / dirCount);
		
		for(int i = 0; i <= dirCount-1; i++)
		{		
			string dir = dirs[i].Replace("\\", "/");
			string[] files = FileManagerSecure.GetFiles(dir);
			
			string lst = dir.Split('/').Last();
			string newd = newDir + lst;
			
			if ( FileManagerSecure.DirectoryExists(newd) == false )
				FileManagerSecure.CreateDirectory(newd);
			
			for(int j = 0; j <= files.Length-1; j++)
			{
				string last = files[j].Split('/').Last();
				string _file = last.Replace("\\", "/");
				
				string oldFile = files[j].Replace("\\", "/");
				string newFile = newDir + _file;
				
				if ( FileManagerSecure.FileExists(newFile) == false )
				{
					FileManagerSecure.CopyFile(oldFile, newFile);
				}
				
				fileCount += 1;
			}	
			
			percent += (int) increment;
			myLabel.text = "Installing... (" + percent.ToString() + ")";	
		}
		
		char symbol = '\u2022';
		myLabel.text = "<b>Pose Manager Library installed</b>\n";
		myLabel.text += symbol.ToString() + " " + dirCount.ToString() + " directory\n";
		myLabel.text += symbol.ToString() + " " + fileCount.ToString() + " files";
	}
	
	// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
	public static string GetPluginPath(MVRScript self)
	{
		string id = self.name.Substring(0, self.name.IndexOf('_'));
		string filename = self.manager.GetJSON()["plugins"][id].Value;
		return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
	}

	#endregion

	#region Relocator

	private IEnumerator SetupRelocator()
	{
		if (creatingSubscene) { yield return 0; }
		creatingSubscene = true;
		
		Atom atom = SuperController.singleton.GetAtomByUid(subname);
		if (atom != null) { SuperController.singleton.RemoveAtom(atom); }
		atom = null;
		
		if (atom == null)
		{
			yield return StartCoroutine(SuperController.singleton.AddAtomByType("Empty", subname));
			atom = SuperController.singleton.GetAtomByUid(subname);	
			atom.transform.position = new Vector3(0f, 0f, 0f);
		}
		
		yield return new WaitForSecondsRealtime(0.1f);
		
		// setup subscene
		for(int i = 1; i <= person.Length-1; i++)
		{
			Atom _atom = null;
					
			if (person[i].val != "")
				_atom = SuperController.singleton.GetAtomByUid(person[i].val);

			if (_atom)
			{
				_atom.parentAtom = atom;
				person[i].val = _atom.name;
			}
		}

		for(int i = 1; i <= obj3ct.Length-1; i++)
		{
			Atom _atom = null;
					
			if (obj3ct[i].val != "")
				_atom = SuperController.singleton.GetAtomByUid(obj3ct[i].val);

			if (_atom)
			{
				_atom.parentAtom = atom;
				obj3ct[i].val = _atom.name;
			}
		}

		creatingSubscene = false;
	}

	#endregion

	#region Move 

	private void MoveToOrigin()
	{
		StartCoroutine( _MoveToOrigin() );
	}
	
	private IEnumerator _MoveToOrigin()
	{
		bool phys = enablePhysics;
		if (phys) { DisablePhysics(); }
		yield return new WaitForSecondsRealtime(0.1f);
		
		Atom atom = SuperController.singleton.GetAtomByUid(subname);

		if (atom != null)
		{
			atom.transform.position = new Vector3(0f, 0f, 0f);
		}
		
		yield return new WaitForSecondsRealtime(0.1f);
		if (phys) { RestorePhysics(); }
	}
	
	private void MoveToCamera()
	{
		StartCoroutine( _MoveToCamera() );
	}
	
	private IEnumerator _MoveToCamera()
	{
		bool phys = enablePhysics;
		if (phys) { DisablePhysics(); }
		yield return new WaitForSecondsRealtime(0.1f);
		
		Atom atom = SuperController.singleton.GetAtomByUid(subname);

		if (atom != null)
		{
			Camera cam = CameraTarget.centerTarget.targetCamera;
			if (cam != null) 
			{
				atom.transform.position = cam.transform.position + cam.transform.TransformDirection(Vector3.forward) * 3.0f;
			}
		}
		
		yield return new WaitForSecondsRealtime(0.1f);
		if (phys) { RestorePhysics(); }
	}
	
	private void Move()
	{
		Atom atom = SuperController.singleton.GetAtomByUid(subname);
		
		if (atom != null)
		{
			FreeControllerV3 ctrl = atom.GetStorableByID("control") as FreeControllerV3;
			
			if (ctrl != null)
				SuperController.singleton.SelectController(ctrl);
		}
	}

	#endregion

	#region Physics

	private void RestorePhysics()
	{
		enablePhysics = true;
		enablePhys.label = "<b>Enable physics</b>";
		disablePhys.label = "Disable physics";
		enablePhys.buttonColor = new Color(0.780f, 0.899f, 0.780f);
		disablePhys.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		
		List<Atom> atoms = SuperController.singleton.GetAtoms();
		
		foreach (Atom atom in atoms)
		{
			atom.SetFreezePhysics(false);
		}
	}
	
	private void DisablePhysics()
	{
		enablePhysics = false;
		enablePhys.label = "Enable physics";
		disablePhys.label = "<b>Disable physics</b>";
		enablePhys.buttonColor = new Color(0.780f, 0.780f, 0.899f);
		disablePhys.buttonColor = new Color(0.255f/0.255f, 0.244f/0.255f, 0.174f/0.255f);
		
		List<Atom> atoms = SuperController.singleton.GetAtoms();
		
		foreach (Atom atom in atoms)
		{
			atom.SetFreezePhysics(true);
		}
	}

	#endregion

	#region Utils

	private void SelectAtom(Atom atom)
	{
		if (atom != null)
		{
			FreeControllerV3 ctrl = atom.GetStorableByID("control") as FreeControllerV3;
				
			if (ctrl != null)
				SuperController.singleton.SelectController(ctrl);
		}
	}

	#endregion

	#region Export Poses

	private void Export()
	{
		SuperController.singleton.ShowMonitorUI();
		
		try
		{
			var fileBrowserUI = SuperController.singleton.fileBrowserUI;
			fileBrowserUI.SetTitle("Save poses (DON'T FORGET TO SELECT THE OBJECTS!!!)");
			fileBrowserUI.fileRemovePrefix = null;
			fileBrowserUI.hideExtension = false;
			fileBrowserUI.keepOpen = false;
			fileBrowserUI.fileFormat = _saveExt;
			fileBrowserUI.defaultPath = _saveFolder + "Library";
			fileBrowserUI.showDirs = true;
			fileBrowserUI.shortCuts = null;
			fileBrowserUI.browseVarFilesAsDirectories = false;
			fileBrowserUI.SetTextEntry(true);
			fileBrowserUI.Show(ExportFileSelected);
			fileBrowserUI.ActivateFileNameField();
		}
		catch (Exception exc)
		{
			SuperController.LogError($"PoseManager: Failed to save file dialog: {exc}");
		}
	}
		
	private void ExportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;
		if (!path.ToLower().EndsWith($".{_saveExt}")) path += $".{_saveExt}";

		try
		{
			jc = new JSONClass();
			jc = SuperController.singleton.GetSaveJSON(null, true, false);
			FillExportJSON(jc);
			
			SuperController.singleton.SaveJSON(jc, path);
			SuperController.singleton.DoSaveScreenshot(path);
			
			myLabel.text = path;
		}
		catch (Exception exc)
		{
			SuperController.LogError($"PoseManager: Failed to export poses: {exc}");
		}
	}

	void FillExportJSON(JSONClass jc)
	{
		List<string> peopleChoices = SuperController.singleton.GetAtoms().Where(a => a.category == "People").Select(a => a.name).ToList();

		int peopleCount = peopleChoices.Count;
		jc.Add("PeopleCount", peopleCount.ToString());
		int count = 0;
			
		foreach (string atomUID in peopleChoices)
		{
			count++;
			Atom _atom = SuperController.singleton.GetAtomByUid(atomUID);
			jc.Add("Person" + count.ToString(), atomUID);
			jc.Add("PType" + count.ToString(), _atom.type);
		}
		
		// ----
		
		List<Atom> otherChoices = new List<Atom>();
		
		for(int i = 1; i <= obj3ct.Length-1; i++)
		{
			if ( obj3ct[i].val != "")
			{
				Atom _atom = SuperController.singleton.GetAtomByUid(obj3ct[i].val);
				
				if (_atom)
				{
					otherChoices.Add(_atom);
				}
			}				
		}
		
		jc.Add("ObjectCount", otherChoices.Count.ToString());
		count = 0;

		foreach (Atom _atom in otherChoices)
		{
			count++;
			jc.Add("Object" + count.ToString(), _atom.name);
			
			if (_atom.type != "")
				jc.Add("OType" + count.ToString(), _atom.type);
			else
				jc.Add("OType" + count.ToString(), "INVALID_TYPE");
		}
	}
	
	#endregion
	
	#region Import Poses
	
	private void Import() // delete or not the objects it spawns
	{
		SuperController.singleton.ShowMonitorUI();
		
		try
		{
			SuperController.singleton.GetMediaPathDialog(ImportFileSelected, _saveExt, _saveFolder + "Library");
		}
		catch (Exception exc)
		{
			SuperController.LogError($"PoseManager: Failed to open file dialog: {exc}");
		}
	}
		
	private void ImportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;

		try
		{
			jc = JSON.Parse(SuperController.singleton.ReadFileIntoString(path)) as JSONClass;
			StartCoroutine( LoadExportedJSON(jc) );
			
			myLabel.text = path;
		}
		catch (Exception exc)
		{
			SuperController.LogError($"PoseManager: Failed to import pose: {exc}");
		}
	}
	
	private JSONClass GetPose(JSONArray atoms, string name)
	{
		for (int i = 0; i < atoms.Count; i++)
		{	
			JSONClass atomObj = atoms[i].AsObject;
			string id = atomObj["id"].Value; // id = "Person", "Person#2", ...
			
			if (id == name)
				return atomObj;
		}
		
		return null;
	}
	
	private IEnumerator LoadExportedJSON(JSONClass jc)
	{
		if ( (jc.HasKey("PeopleCount") == false) || (jc.HasKey("ObjectCount") == false)  )
		{
			SuperController.LogMessage( "Invalid Pose Manager File!" );
			yield return 0;
		}
		
		Atom atom = SuperController.singleton.GetAtomByUid(subname);
		
		if (atom != null)
		{
			atom.transform.position = new Vector3(0f, 0f, 0f); // why this doesn't work
		}
		
		atomCalls.Clean(true);
		
		List<string> peopleChoices = SuperController.singleton.GetAtoms().Where(a => a.category == "People").Select(a => a.name).ToList();
				
		JSONArray atoms = jc["atoms"].AsArray;
		
		// ---------------------------------------------------------------------------------------
		// people setup
		// ---------------------------------------------------------------------------------------
	
		int peopleCount = int.Parse(jc["PeopleCount"].Value);
		AddMissingPersons(peopleCount);
		
		atomCalls.p_index = 0;
		
		for(int i = 1; i <= peopleCount; i++)
		{
			// --------------------------------------
			// get a person from the file
			// --------------------------------------
			
			string name = jc["Person" + i.ToString()].Value;
			JSONClass atomObj = GetPose(atoms, name);

			// --------------------------------------
			// get a person from the world
			// --------------------------------------
			
			Atom _atom = null;
			
			// try to select a person selected by the user
			if (i <= person.Length-1)
			{ 
				if (person[i].val != "")
				_atom = SuperController.singleton.GetAtomByUid(person[i].val);
			}
			
			// try to select an existing person
			if (_atom == null)
			{
				if (peopleChoices.Count > 0)
					_atom = SuperController.singleton.GetAtomByUid(peopleChoices[0]);
			}
			
			// --------------------------------------
			// process that person or create one
			// --------------------------------------
			
			if (_atom == null)
			{
				string nn = GetRandomName("Person");
				yield return StartCoroutine(SuperController.singleton.AddAtomByType("Person", nn));
				_atom = SuperController.singleton.GetAtomByUid(nn);	
				atomCalls.AddAtom(_atom, atomObj);
			}
			else
			{
				peopleChoices.Remove(_atom.name);
				atomCalls.AddAtom(_atom, atomObj);
			}
			
			_atom.SetFreezePhysics(true);
		}
		
		// ---------------------------------------------------------------------------------------
		// object setup
		// ---------------------------------------------------------------------------------------
		
		for(int i = 1; i <= obj3ct.Length-1; i++)
		{
			obj3ct[i].val = "";
		}
		
		atomCalls.o_index = 0;
		
		List<string> otherChoices = SuperController.singleton.GetAtoms().Where(a => a.category != "People").Select(a => a.name).ToList();
		
		int objectCount = int.Parse(jc["ObjectCount"].Value);
		if (spawnObjects.val == false) { objectCount = 0; }
		
		AddMissingObjects(objectCount);
		
		for(int i = 1; i <= objectCount; i++)
		{
			// --------------------------------------
			// get the object from the file
			// --------------------------------------
			
			string name = jc["Object" + i.ToString()].Value;
			JSONClass atomObj = GetPose(atoms, name);
			
			string type = jc["OType" + i.ToString()].Value;
			
			// --------------------------------------
			// get the object from the world
			// --------------------------------------
			
			Atom _atom = null;
			
			foreach (string atomUID in otherChoices)
			{
				Atom _atm = SuperController.singleton.GetAtomByUid(atomUID);
				
				if (_atm.type == type)
				{
					_atom = _atm;
					break;
				}
			}
			
			// --------------------------------------
			// process that object or create one
			// --------------------------------------
			
			if (_atom == null)
			{
				string nn = GetRandomName(type);
				yield return StartCoroutine(SuperController.singleton.AddAtomByType(type, nn));
				_atom = SuperController.singleton.GetAtomByUid(nn);	
				atomCalls.AddAtom(_atom, atomObj);
			}
			else
			{
				otherChoices.Remove(_atom.name);
				atomCalls.AddAtom(_atom, atomObj);
			}

			_atom.SetFreezePhysics(true);
		}
		
		yield return new WaitForSecondsRealtime(2.5f);
		
		atomCalls.RestoreAll();
		
		if (deleteObjects.val == false)
			deleteObjects.val = true;
		
		StartCoroutine( SetupRelocator() );
	}
	
	private string GetRandomName(string name)
	{
		return name + UnityEngine.Random.Range(1000, 10000000).ToString();
	}

	#endregion

	#region Restore
	
	public class AtomCallbacks
	{
		public List<Atom> delNames; // atoms to delete
		private PoseManager parent;
		public List<atomSlot> slots;
		public int p_index;
		public int o_index;
        
		// contructor
		public AtomCallbacks()
		{
			slots = new List<atomSlot>();
			delNames = new List<Atom>();
		}
		
		public void SetParent(PoseManager _parent)
		{
			parent = _parent;
		}
		
		public struct atomSlot
		{
			public Atom atom;
			public JSONClass jc;
		   
			//Constructor (not necessary, but helpful)
			public atomSlot(Atom atom, JSONClass jc)
			{
				this.atom = atom;
				this.jc = jc;
			}
		}
		
		public void RestoreAll()
		{
			foreach (atomSlot slot in slots)
			{
				RestoreAtom(slot);
			}
		}
		
		private void RestoreAtom(atomSlot slot)
		{
			if (slot.atom.category == "People")
			{
				p_index++;
				parent.person[p_index].val = slot.atom.name;
				
				slot.atom.PreRestore(restorePhysical: true, restoreAppearance: false);
				slot.atom.RestoreTransform(slot.jc);
				slot.atom.Restore(slot.jc, true, false, false);
				slot.atom.LateRestore(slot.jc, true, false, false);
				slot.atom.PostRestore(restorePhysical: true, restoreAppearance: false);
				slot.atom.ResetPhysics(fullReset: false);
			}
			else
			{
				o_index++;
				parent.obj3ct[o_index].val = slot.atom.name;
				
				slot.atom.PreRestore(restorePhysical: true, restoreAppearance: true);
				slot.atom.RestoreTransform(slot.jc);
				slot.atom.Restore(slot.jc, true, true, true);
				slot.atom.LateRestore(slot.jc, true, true, true);
				slot.atom.PostRestore(restorePhysical: true, restoreAppearance: true);
				slot.atom.ResetPhysics(fullReset: false);
			}	

			if (parent.enablePhysics)
				slot.atom.SetFreezePhysics(false);
			else
				slot.atom.SetFreezePhysics(true);			
		}
		
		public void AddAtom(Atom atom, JSONClass _jc)
		{
			slots.Add( new atomSlot(atom, _jc) );
			
			if (atom.category != "People")
			{
				if (parent.deleteObjects.val)
					delNames.Add(atom);
			}
		}
		
		public void Clean(bool clearall)
		{
			if (clearall)
			{
				foreach (Atom atom in delNames)
				{
					SuperController.singleton.RemoveAtom(atom);
				}
			}
			
			slots.Clear();
			delNames.Clear();
		}
	}

	private void AddNewAtom(string atomName)
	{
		SuperController.singleton.AddAtomByType(atomName, false, false, false);
	}
	
	public void DeleteAtom(string atomName)
	{
		Atom _atom = SuperController.singleton.GetAtomByUid(atomName);
		if (_atom) { SuperController.singleton.RemoveAtom(_atom); }	
	}		
	
	#endregion

	#region Update

	public void Update()
	{

		/*

		// FOR TESTING PURPOSES ONLY
		if (Input.GetKeyDown("5"))
		{
			
			string name = "";
			
			Atom cur = SuperController.singleton.GetSelectedAtom();
			
			if (cur == null)
			{ 
				name = "Person";
			}
			else
			{
				if (cur.name == "Person")
					name = "Guy";
				else
					name = "Person";
			}
			
			Atom atom = SuperController.singleton.GetAtomByUid(name);
			
			if (atom != null)
			{
				FreeControllerV3 ctrl = atom.GetStorableByID("hipControl") as FreeControllerV3;
				
				if (ctrl != null)
					SuperController.singleton.SelectController(ctrl);
			}
		}
		
		if (Input.GetKeyDown("6"))
		{
			string name = "";
			
			FreeControllerV3 ctrl = SuperController.singleton.GetSelectedController();
			
			if (ctrl == null)
			{ 
				name = "penisBaseControl";
			}
			else
			{
				if (ctrl.name == "penisBaseControl")
					name = "penisTipControl";
				else
					name = "penisBaseControl";
			}
			
			Atom atom = SuperController.singleton.GetAtomByUid("Guy");
			
			if (atom != null)
			{
				ctrl = atom.GetStorableByID(name) as FreeControllerV3;
				
				if (ctrl != null)
					SuperController.singleton.SelectController(ctrl);
			}
		}


		// FOR TESTING PURPOSES ONLY
		if (Input.GetKeyDown("2"))
		{
			Atom cur = SuperController.singleton.GetSelectedAtom();
			if (cur == null) { return; }

			if (cur != null)
			{
				FreeControllerV3 ctrl = cur.GetStorableByID("control") as FreeControllerV3;

				if (cur.type == "ISCylinder")
				{
					nposition = ctrl.transform.position;
					nrotation = ctrl.transform.rotation;
				}	
			}
		}

		if (Input.GetKeyDown("3"))
		{
			List<string> objects = getTypeChoices("ISCylinder");
			visibles = !visibles;

			foreach (string atomUID in objects)
			{
				Atom atom = SuperController.singleton.GetAtomByUid(atomUID);
				atom.SetOn(visibles);
			}
		}

		if (Input.GetKeyDown("4"))
		{
			Atom cur = SuperController.singleton.GetSelectedAtom();
			if (cur == null) { return; }

			if (cur != null)
			{
				FreeControllerV3 ctrl = cur.GetStorableByID("control") as FreeControllerV3;

				if (cur.type == "ISCylinder")
				{
					ctrl.transform.position = nposition;
					ctrl.transform.rotation = nrotation;
				}
			}
		}

		*/

	}

	/*
	private Vector3 nposition;
	private Quaternion nrotation;
	private bool visibles = true;
	*/

	// get all instances of one type
	public static List<string> getTypeChoices(string _type)
	{
		List<string> choices = new List<string>();

		foreach (Atom a in SuperController.singleton.GetAtoms())
		{
			if (a.type == _type)
				choices.Add(a.name);
		}

		return choices;
	}

	#endregion

} // End class

